import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
from kmodes.kprototypes import matching_dissim, euclidean_dissim
from sklearn.metrics import silhouette_score
from typing import Literal


class NClusterSearch:
    """
Optimal Number of Clusters Searcher
====================================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/cluster.html#optimal-number-of-clusters-searcher).

Description
-----------

A helper class to identify the optimal number of clusters for clustering algorithms.

This class offers methods like the elbow method and silhouette method to evaluate
and suggest the best number of clusters for the provided data. The class currently
supports KMeans, KMedoids, KModes, and KPrototypes as estimators.

Parameters:
-----------
`estimator` : instance of clustering model
    The clustering algorithm for which you want to find the optimal number of clusters.
    Supported estimators: `KMeans`, `KMedoids`, `KModes`, `KPrototypes`.

`method` : str, optional (default = 'elbow')
    The method used to determine the optimal number of clusters.
    Accepted values: 'elbow', 'silhouette'.

`min_clusters` : int, optional (default = 1)
    The minimum number of clusters to consider.

`max_clusters` : int, optional (default = 10)
    The maximum number of clusters to consider.

`step` : int, optional (default = 1)
    The step size for increasing the number of clusters during the search.

`random_state` : int, optional (default = 123)
    Random seed for reproducibility.

`verbose` : bool, optional (default = False)
    If True, the process will print details as it proceeds.

Examples:
---------

Example 1: Using the elbow method with KMeans

>>> from sklearn.cluster import KMeans
>>> from sklean.datasets import load_iris
>>> data = load_iris().data
>>> new_data = load_iris().data[:10]
>>> searcher = NClusterSearch(estimator=KMeans(), method='elbow')
>>> searcher.fit(data)
>>> searcher.plot()
>>> labels = searcher.labels_
>>> predicted_labels = searcher.predict(new_data)
>>> optimal_model = searcher.best_estimator_
>>> optimal_clusters = searcher.optimal_clusters_


Example 2: Using KModes with custom arguments

>>> from kmodes.kmodes import KModes
>>> import pandas as pd
>>> import numpy as np
>>> np.random.seed(2023)
>>> data = pd.DataFrame(np.random.randint(0, 10, size=(100, 4))).apply(lambda x: x.astype('object'))
>>> kmodes = KModes(init='Huang', n_init=5, verbose=1)
>>> searcher = NClusterSearch(estimator=kmodes, method='elbow')
>>> searcher.fit(data)

Example 3: Using the silhouette method with KMedoids

>>> from sklean.datasets import load_iris
>>> from sklearn_extra.cluster import KMedoids
>>> data = load_iris().data
>>> searcher = NClusterSearch(estimator=KMedoids(), method='silhouette')
>>> searcher.fit(data)
>>> searcher.plot()

Example 4: Using the silhouette method with KPrototypes

>>> import pandas as pd
>>> import numpy as np
>>> from kmodes.kprototypes import KPrototypes
>>> np.random.seed(2023)
>>> data = pd.DataFrame(np.random.randint(0, 10, size=(100, 4))).apply(lambda x: x.astype('object'))
>>> data["new"] = np.random.randint(0, 10, size=(100, 1))
>>> searcher = NClusterSearch(estimator=KPrototypes(), method='silhouette', verbose=True)
>>> searcher.fit(data, categorical=[0, 1, 2, 3])
>>> searcher.plot()

Example 5: Using the silhouette method with KPrototypes and custom alpha for distance calculation

>>> import pandas as pd
>>> import numpy as np
>>> from kmodes.kprototypes import KPrototypes
>>> np.random.seed(2023)
>>> data = pd.DataFrame(np.random.randint(0, 10, size=(100, 4))).apply(lambda x: x.astype('object'))
>>> data["new"] = np.random.randint(0, 10, size=(100, 1))
>>> searcher = NClusterSearch(estimator=KPrototypes(), method='silhouette')
>>> searcher.fit(data, categorical=[0, 1, 2, 3], alpha=0.1)
>>> searcher.plot()
    """

    def __init__(
        self,
        estimator,
        method: Literal["elbow", "silhouette"] = "elbow",
        *,
        min_clusters: int = 1,
        max_clusters: int = 10,
        step: int = 1,
        random_state: int = 123,
        verbose: bool = False,
    ):
        self.estimator = estimator
        self.method = method
        self.min_clusters = min_clusters
        if self.method == "silhouette" and (self.min_clusters == 1):
            self.min_clusters = 2
        self.max_clusters = max_clusters
        self.step = step
        self.random_state = random_state
        self.__range = list(range(self.min_clusters, self.max_clusters + 1, self.step))
        self.verbose = verbose

        if type(self.estimator).__name__ in ["KMeans", "KMedoids"]:
            self.__cost = "inertia_"
        elif type(self.estimator).__name__ in ["KModes", "KPrototypes"]:
            self.__cost = "cost_"
        else:
            raise Exception(
                f"The estimator '{type(self.estimator).__name__}' is not implemented yet."
            )

        if self.min_clusters > self.max_clusters:
            raise Exception(
                f"The minimum number of clusters ({self.min_clusters}) is greater than the maximum number of clusters ({self.max_clusters})."
            )

        if self.estimator.__dict__.get("random_state") is None:
            self.estimator.set_params(random_state=self.random_state)

        if self.method not in ["elbow", "silhouette"]:
            raise Exception(f"The method '{self.method}' is not implemented yet.")

    def fit(self, X, **kwargs):
        """
Fits the estimator with the data over a range of cluster numbers.

Parameters:
-----------
`X` : array-like, shape (n_samples, n_features)
    The data to determine the optimal number of clusters for.
    
`**kwargs` : Additional keyword arguments to be passed to the estimator's fit method.

Returns:
--------
`self` : object
    Returns an instance of `NClusterSearch()`.

Example:
--------

>>> from sklearn.cluster import KMeans
>>> from sklean.datasets import load_iris
>>> data = load_iris().data
>>> searcher = NClusterSearch(estimator=KMeans(), method='elbow')
>>> searcher.fit(data)
        """
        self.__X = X

        if self.__X.shape[0] < self.max_clusters:
            raise Exception(
                f"The number of samples ({self.__X.shape[0]}) is lower than the maximum number of clusters ({self.max_clusters})."
            )
        if self.__X.shape[0] < self.min_clusters:
            raise Exception(
                f"The number of samples ({self.__X.shape[0]}) is lower than the minimum number of clusters ({self.max_clusters})."
            )

        self.__kwargs = kwargs

        if self.method == "elbow":
            self.__elbow_method()
        elif self.method == "silhouette":
            self.__silhuette_method()
        self.__fitted = True

        return self

    @property
    def best_estimator_(self):
        """
Returns the estimator with the optimal number of clusters.

This property can be accessed only after calling the fit method.

Returns:
--------
best_estimator : instance of clustering model
    Estimator set to the optimal number of clusters and trained with the data.

Example:
--------

>>> from sklearn.cluster import KMeans
>>> from sklean.datasets import load_iris
>>> data = load_iris().data
>>> model = NClusterSearch(estimator=KMeans(), method='elbow')
>>> model.fit(data)
>>> optimal_model = model.best_estimator_
        """
        if not self.__fitted:
            raise Exception(
                "The model is not fitted yet. Call 'fit' with appropriate arguments before using this property."
            )
        return self.estimator.set_params(n_clusters=self.optimal_clusters_).fit(
            self.__X, **self.__kwargs
        )

    def __compute_elbow_point(self, sse):
        # Calculate the line from first to last point
        x = np.arange(len(sse))
        y = np.array(sse)
        line_k = (y[-1] - y[0]) / (x[-1] - x[0])
        line_b = y[0] - line_k * x[0]

        # Calculate distance from each point to the line
        line_y = line_k * x + line_b
        distance = np.abs(line_y - y) / np.sqrt(1 + line_k**2)

        # Return the index of point with max distance
        return np.argmax(distance)

    def __elbow_method(self):
        self.__costs = []
        iterable = tqdm(self.__range) if self.verbose else self.__range
        if self.verbose:
            print("Calculating costs...")
        for n_clusters in iterable:
            self.estimator.set_params(n_clusters=n_clusters)
            self.estimator.fit(self.__X, **self.__kwargs)
            self.__costs.append(self.estimator.__getattribute__(self.__cost))

        self.optimal_clusters_ = self.__range[self.__compute_elbow_point(self.__costs)]
        self.estimator.n_clusters = self.optimal_clusters_

    def __create_dm__kmodes(self, dataset):
        if type(dataset).__name__ == "DataFrame":
            dataset = dataset.values
        lenDataset = len(dataset)
        if self.verbose:
            print("Creating distance matrix...")
        distance_matrix = np.zeros(lenDataset * lenDataset).reshape(
            lenDataset, lenDataset
        )
        for i in range(lenDataset):
            for j in range(lenDataset):
                x1 = dataset[i].reshape(1, -1)
                x2 = dataset[j].reshape(1, -1)
                distance = matching_dissim(x1, x2)
                distance_matrix[i][j] = distance[0]
                distance_matrix[j][i] = distance[0]
        return distance_matrix

    def __mixed_distance__kproto(self, a, b, alpha=0.01):
        if self.__kwargs["categorical"] is None:
            num_score = euclidean_dissim(a, b)
            return num_score
        else:
            cat_index = self.__kwargs["categorical"]
            a_cat = []
            b_cat = []
            for index in cat_index:
                a_cat.append(a[index])
                b_cat.append(b[index])
            a_num = []
            b_num = []
            l = len(a)
            for index in range(l):
                if index not in cat_index:
                    a_num.append(a[index])
                    b_num.append(b[index])

            a_cat = np.array(a_cat).reshape(1, -1)
            a_num = np.array(a_num).reshape(1, -1)
            b_cat = np.array(b_cat).reshape(1, -1)
            b_num = np.array(b_num).reshape(1, -1)
            cat_score = matching_dissim(a_cat, b_cat)
            num_score = euclidean_dissim(a_num, b_num)
            return cat_score + num_score * alpha

    def __create_dm__kproto(self, dataset, alpha=0.1):
        if type(dataset).__name__ == "DataFrame":
            dataset = dataset.values
        lenDataset = len(dataset)
        if self.verbose:
            print("Creating distance matrix...")
        distance_matrix = np.zeros(lenDataset * lenDataset).reshape(
            lenDataset, lenDataset
        )
        for i in range(lenDataset):
            for j in range(lenDataset):
                x1 = dataset[i]
                x2 = dataset[j]
                distance = self.__mixed_distance__kproto(x1, x2, alpha=alpha)
                distance_matrix[i, j] = distance[0]
                distance_matrix[j, i] = distance[0]
        return distance_matrix

    def __silhuette_method(self):
        self.__sils = []
        iterable = tqdm(self.__range) if self.verbose else self.__range
        if self.verbose:
            print("Calculating silhouette scores...")
        if type(self.estimator).__name__ in ["KMeans", "KMedoids"]:
            for n_clusters in iterable:
                self.estimator.n_clusters = n_clusters
                self.estimator.fit(self.__X, **self.__kwargs)
                self.__sils.append(silhouette_score(self.__X, self.estimator.labels_))
        elif type(self.estimator).__name__ == "KModes":
            distance_matrix = self.__create_dm__kmodes(self.__X)
            for n_clusters in iterable:
                self.estimator.n_clusters = n_clusters
                self.estimator.fit(self.__X, **self.__kwargs)
                cluster_labels = self.estimator.labels_
                score = silhouette_score(
                    distance_matrix, cluster_labels, metric="precomputed"
                )
                self.__sils.append(score)
        elif type(self.estimator).__name__ == "KPrototypes":
            if not self.__kwargs.get("categorical"):
                raise Exception(
                    "The argument 'categorical' is required for the KPrototypes estimator."
                )
            if not isinstance(self.__kwargs["categorical"], list):
                raise Exception("The argument 'categorical' must be a list.")
            if not self.__kwargs.get("alpha"):
                self.__alpha = 0.01
            else:
                self.__alpha = self.__kwargs["alpha"]
                del self.__kwargs["alpha"]
            distance_matrix = self.__create_dm__kproto(self.__X, alpha=self.__alpha)
            for n_clusters in iterable:
                self.estimator.n_clusters = n_clusters
                self.estimator.fit(self.__X, **self.__kwargs)
                cluster_labels = self.estimator.labels_
                score = silhouette_score(
                    distance_matrix, cluster_labels, metric="precomputed"
                )
                self.__sils.append(score)
        else:
            raise Exception(
                f"The estimator '{type(self.estimator).__name__}' is not implemented yet."
            )

        self.optimal_clusters_ = self.__range[np.argmax(self.__sils)]
        self.estimator.n_clusters = self.optimal_clusters_

    def __plot_elbow(self, ax):
        if ax is None:
            plt.figure(figsize=(10, 6))
            plt.plot(self.__range, self.__costs, "-o")
            plt.xlabel("Number of clusters")
            plt.ylabel("Cost")
            plt.title("Elbow Method For Optimal Cluster Number")
            plt.axvline(
                x=self.optimal_clusters_,
                color="r",
                linestyle="--",
                label=f"Elbow point = {self.optimal_clusters_}",
            )
            plt.legend()
        else:
            ax.plot(self.__range, self.__costs, "-o")
            ax.set_xlabel("Number of clusters")
            ax.set_ylabel("Cost")
            ax.set_title("Elbow Method For Optimal Cluster Number")
            ax.axvline(
                x=self.optimal_clusters_,
                color="r",
                linestyle="--",
                label=f"Elbow point = {self.optimal_clusters_}",
            )
            ax.legend()

    def __plot_silhouette(self, ax):
        if ax is None:
            plt.figure(figsize=(10, 6))
            plt.plot(self.__range, self.__sils, "-o")
            plt.xlabel("Number of clusters")
            plt.ylabel("Silhouette Score")
            plt.title("Silhouette Method For Optimal Cluster Number")
            plt.axvline(
                x=self.optimal_clusters_,
                color="r",
                linestyle="--",
                label=f"Maximum silhouette point = {self.optimal_clusters_}",
            )
            plt.legend()
        else:
            ax.plot(self.__range, self.__sils, "-o")
            ax.set_xlabel("Number of clusters")
            ax.set_ylabel("Silhouette Score")
            ax.set_title("Silhouette Method For Optimal Cluster Number")
            ax.axvline(
                x=self.optimal_clusters_,
                color="r",
                linestyle="--",
                label=f"Maximum silhouette point = {self.optimal_clusters_}",
            )
            ax.legend()

    def plot(self, *, ax=None):
        """
Plots the results of the selected method (either 'elbow' or 'silhouette') to visualize the optimal cluster number.

Parameters:
-----------
ax : matplotlib axis object, optional
    Axis on which to draw the plot. If None, a new figure and axis will be created.

Returns:
--------
None

Example:
--------

>>> from sklearn.cluster import KMeans
>>> from sklean.datasets import load_iris
>>> data = load_iris().data
>>> model = NClusterSearch(estimator=KMeans(), method='elbow')
>>> model.fit(data)
>>> model.plot()
        """
        if not self.__fitted:
            raise Exception(
                "The model is not fitted yet. Call 'fit' with appropriate arguments before using this method."
            )
        if ax is not None and not isinstance(ax, plt.Axes):
            raise Exception("The argument 'ax' must be a matplotlib Axes object.")
        if self.method == "elbow":
            self.__plot_elbow(ax=ax)
        elif self.method == "silhouette":
            self.__plot_silhouette(ax=ax)

    def predict(self, X, **kwargs):
        """
Predicts the closest cluster for each sample in `X` using the `best_estimator_`.

Parameters:
-----------
`X` : array-like, shape (n_samples, n_features)
    New data to predict cluster labels.
    
`**kwargs` : Additional keyword arguments to be passed to the estimator's predict method.

Returns:
--------
`labels` : array, shape (n_samples,)
    Index of the cluster each sample belongs to.

Example:
--------

>>> from sklearn.cluster import KMeans
>>> from sklean.datasets import load_iris
>>> data = load_iris().data
>>> model = NClusterSearch(estimator=KMeans(), method='elbow')
>>> model.fit(data)
>>> predictions = model.predict(new_data)
        """
        if not self.__fitted:
            raise Exception(
                "The model is not fitted yet. Call 'fit' with appropriate arguments before using this method."
            )
        
        return self.best_estimator_.predict(X, **kwargs)

    @property
    def labels_(self):
        """
Returns the labels of each point for the best estimator.

This property can be accessed only after calling the fit method.

Returns:
--------
labels : array, shape (n_samples,)
    Index of the cluster each sample belongs to.

Example:
--------

>>> from sklearn.cluster import KMeans
>>> from sklean.datasets import load_iris
>>> data = load_iris().data
>>> model = NClusterSearch(estimator=KMeans(), method='elbow')
>>> model.fit(data)
>>> labels = model.labels_
        """
        return self.best_estimator_.labels_

    def __repr__(self) -> str:
        return f"NClusterSearch(estimator={self.estimator}, method='{self.method}', min_clusters={self.min_clusters}, max_clusters={self.max_clusters}, step={self.step}, random_state={self.random_state})"



from typing import Any, Dict, Literal, Union

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame, Series, concat
from patsy import build_design_matrices, dmatrices
from scipy.stats import norm
from sklearn.linear_model import LogisticRegression as LogisticRegression_
from statsmodels.api import GLM, OLS, Logit

from estyp.testing.__base import __nested_models_test as nested_models_test
from estyp.testing.__base import bcolors


class LogisticRegression(LogisticRegression_):
    """
    Logistic Regression
    ===================

    View this in the [online documentation](https://estyp.readthedocs.io/en/latest/linear_model.html#logistic-regression).

    Description
    -----------

    This class implements a logistic regression model. It inherits from the `sklearn.linear_model.LogisticRegression` class, but adds additional methods for calculating confidence intervals, p-values, and model summaries.

    Parameters
    ----------

    - `X`: A Pandas DataFrame or a NumPy array containing the model predictors.
    - `y`: A Pandas Series or a NumPy array containing the model response.
    - `penalty`: The type of penalty to use. Can be one of `"none"` (default). `"l1"`, `"l2"`, or `"elasticnet"`.
    - `dual`: Whether to use the dual formulation of the problem.
    - `tol`: The tolerance for convergence.
    - `C`: The regularization strength.
    - `fit_intercept`: Whether to fit an intercept term.
    - `intercept_scaling`: The scaling factor for the intercept term.
    - `class_weight`: None (default), "balanced" or a dictionary that maps class labels to weights.
    - `random_state`: The random seed.
    - `solver`: The solver to use. Can be one of `"lbfgs"` (default), `"liblinear"`, `"newton-cg"`, `"newton-cholesky"`, `"sag"`, or `"saga"`.
    - `max_iter`: The maximum number of iterations.
    - `multi_class`: The type of multi-class classification to use. Can be one of `"auto"`, `"ovr"`, or `"multinomial"`.
    - `verbose`: The verbosity level.
    - `warm_start`: Whether to use the warm start.
    - `n_jobs`: The number of jobs to use for parallel processing.
    - `l1_ratio`: The l1_ratio parameter for elasticnet regularization.

    Properties
    ----------

    - `params`: Returns the estimated values for model parameters.
    - `aic`: Calculates the Akaike information criterion (AIC) for the model.
    - `bic`: Calculates the Bayesian information criterion (BIC) for the model.
    - `cov_matrix`: Returns the covariance matrix for model parametres.
    - `deviance`: Returns the deviance of the model.
    - `deviance_residuals`: Returns the deviance residuals.

    Methods
    -------

    - `fit()`: Fits the model to the data.
    - `predict()`: Predicts the class labels for new data.
    - `conf_int()`: Calculates the confidence intervals for the model coefficients.
    - `se()`: Calculates the standard errors for the model coefficients.
    - `z_values()`: Calculates the z-scores for the model coefficients.
    - `p_values()`: Calculates the p-values for the model coefficients.
    - `summary()`: Prints a summary of the model.

    Examples
    --------

    >>> import numpy as np
    >>> import pandas as pd
    >>> np.random.seed(123)
    >>> data = pd.DataFrame({
    ...     "y": np.random.randint(2, size=100),
    ...     "x1": np.random.uniform(-1, 1, size=100),
    ...     "x2": np.random.uniform(-1, 1, size=100),
    ... })
    >>> formula = "y ~ x1 + x2"
    >>> spec = LogisticRegression.from_formula(formula, data)
    >>> model = spec.fit()
    >>> print(model.summary())
    """

    def __init__(
        self,
        X: Union[DataFrame, np.ndarray] = None,
        y: Union[Series, np.ndarray] = None,
        penalty: Literal["l1", "l2", "elasticnet"] = None,
        *,
        dual=False,
        tol=0.0001,
        C=1,
        fit_intercept=False,
        intercept_scaling=1,
        class_weight=None,
        random_state=2023,
        solver: Literal[
            "lbfgs", "liblinear", "newton-cg", "newton-cholesky", "sag", "saga"
        ] = "lbfgs",
        max_iter=100,
        multi_class: Literal["auto", "ovr", "multinomial"] = "auto",
        verbose=0,
        warm_start: bool = False,
        n_jobs=-1,
        l1_ratio=None,
    ):
        super().__init__(
            penalty=penalty,
            dual=dual,
            tol=tol,
            C=C,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            class_weight=class_weight,
            random_state=random_state,
            solver=solver,
            max_iter=max_iter,
            multi_class=multi_class,
            verbose=verbose,
            warm_start=warm_start,
            n_jobs=n_jobs,
            l1_ratio=l1_ratio,
        )

        if self.__dict__.get("X") is None and self.__dict__.get("y") is None:
            if isinstance(X, np.ndarray):
                X = DataFrame(X, columns=[f"var_{i+1}" for i in range(X.shape[1])])
            if X is not None and y is not None:
                data = concat([X, Series(y, name="y")], axis=1)
                formula = (
                    f"y ~ {' + '.join(X.columns.tolist()) if X.columns.tolist() else 1}"
                )
                y, X = dmatrices(formula, data)
                self.X = X
                self.y = y.reshape(-1)

    @classmethod
    def from_formula(cls, formula, data, **kwargs):
        if "X" in kwargs.keys() or "y" in kwargs.keys():
            raise ValueError(
                "The 'X' and 'y' arguments cannot be used with the 'from_formula' method."
            )
        y, X = dmatrices(formula, data)
        self_ = cls(**kwargs)
        self_.y = y.reshape(-1)
        self_.X = X
        self_.formula = formula
        return self_

    def __log_likelihood(self):
        z = np.dot(self.X, self.coef_.flatten())
        log_likelihood = np.sum(self.y * z - np.log(1 + np.exp(z)))
        return log_likelihood

    def fit(self, disp=None):
        super().fit(self.X, self.y)
        return self

    def conf_int(self, conf_level=0.95):
        z_score = norm.ppf(1 - (1 - conf_level) / 2)
        standard_errors = np.sqrt(np.diag(self.cov_matrix))
        lower_limits = self.coef_[0] - z_score * standard_errors
        upper_limits = self.coef_[0] + z_score * standard_errors
        ci = DataFrame(
            {"[Lower,": lower_limits, "Upper]": upper_limits},
            index=self.X.design_info.column_names,
        )
        return ci

    def se(self):
        return Series(
            np.sqrt(np.diag(self.cov_matrix)),
            index=self.X.design_info.column_names,
            name="S.E.",
        )

    def z_values(self):
        z_est = self.coef_[0] / self.se()
        return Series(z_est, index=self.X.design_info.column_names, name="z")

    def p_values(self):
        p_values = 2 * norm.cdf(-np.abs(self.z_values()))
        return Series(p_values, index=self.X.design_info.column_names, name="Pr(>|z|)")

    def summary(self, conf_level=0.95):
        summary = concat(
            objs=[
                self.params,
                self.se(),
                self.z_values(),
                self.p_values(),
                self.conf_int(conf_level),
            ],
            axis=1,
        )
        return summary

    def predict(self, new_data: DataFrame):
        dsg = build_design_matrices([self.X.design_info], new_data)[0].view()
        return super().predict_proba(dsg)[:, 1]

    @property
    def cov_matrix(self):
        p = self.predict_proba(self.X)[:, 1]
        hess_matrix = np.dot(self.X.T, np.dot(np.diag(p * (1 - p)), self.X))
        cov_matrix = np.linalg.inv(hess_matrix)
        return cov_matrix

    @property
    def params(self):
        return Series(
            self.coef_.flatten().tolist(),
            index=self.X.design_info.column_names,
            name="Estimate",
        )

    @property
    def aic(self):
        aic = -2 * self.__log_likelihood() + 2 * (self.coef_.shape[0] - 1)
        return aic.item()

    @property
    def bic(self):
        bic = -2 * self.__log_likelihood() + np.log(self.X.shape[0]) * (
            self.coef_.shape[0] - 1
        )
        return bic.item()

    @property
    def deviance_residuals(self):
        p = self.predict_proba(self.X)[:, 1]
        residuals = -2 * (self.y * np.log(p) + (1 - self.y) * np.log(1 - p))
        return residuals

    @property
    def deviance(self):
        return np.sum(self.deviance_residuals).item()

    def __repr__(self) -> str:
        return "LogisticRegression()"


class Stepwise:
    """
    Stepwise Selection for Linear Models
    ======================================
    
    View this in the [online documentation](https://estyp.readthedocs.io/en/latest/linear_model.html#stepwise-selection-for-linear-models).
    
    Description
    -----------

    The `Stepwise` class provides a method to perform stepwise model selection, which is a method to add or remove predictors based on their significance, AIC or BIC in a model.

    Parameters:
    -----------
    `formula` : str
        A string representing the formula, using the `patsy` formula syntax.
        For instance, the formula "y ~ x1 + x2" denotes `y` as the dependent variable and `x1` and `x2` as independent variables.

    `data` : DataFrame
        A pandas DataFrame that contains the data for both the dependent and independent variables.

    `model` : Union[GLM, OLS, Logit, LogisticRegression], optional (default = OLS)
        Specifies the type of model to be used. The options are:
        - `GLM`: Generalized Linear Model
        - `OLS`: Ordinary Least Squares Regression
        - `Logit`: Logistic Regression (for binary classification)
        - `LogisticRegression`: Logistic Regression (from the `estyp.linear_model` module)

    `direction` : Literal["both", "forward", "backward"], optional (default = "forward")
        Specifies the direction of the stepwise process:
        - "forward": Start with no predictors and add predictors one-by-one.
        - "backward": Start with all predictors and remove predictors one-by-one.
        - "both": Combination of forward and backward methods.

    `criterion` : Literal["aic", "bic", "f-test"], optional (default = "aic")
        The criterion to be used for adding or removing predictors:
        - "aic": Akaike Information Criterion
        - "bic": Bayesian Information Criterion
        - "f-test": F-test (only available for "forward" and "backward" directions)

    `alpha` : float, optional (default = 0.05)
        The significance level for adding or removing predictors.
        It must be a value between 0 and 1.

    `max_iter` : int, optional (default = 100)
        The maximum number of iterations for the both direction process.

    `formula_params` : Dict[str, Any], optional (default = {})
        Additional parameters to be passed to the model's `from_formula` method.

    `fit_params` : Dict[str, Any], optional (default = {"disp": 0})
        Additional parameters to be passed to the model's `fit` method.

    `verbose` : bool, optional (default = True)
        If set to `False`, the class will not print information about the stepwise process.

    Attributes:
    -----------
    `optimal_model_`  : Model instance
        The optimal model obtained after the stepwise process.

    `optimal_formula_` : str
        The optimal model formula after the stepwise process.

    `optimal_variables_` : list
        List of optimal predictor variables in the final model.

    `optimal_metric_` : float
        The optimal value of the chosen criterion (e.g., AIC, BIC, or F-test) for the final model.

    Methods:
    --------
    `fit()` :
        Conducts the stepwise process based on the specified direction and criterion.

    `plot_history(ax=None)`:
        Plots the history of the chosen criterion during the stepwise.

        Parameters:
        - `ax  : matplotlib.axes.Axes (optional)
            An Axes instance for the plot. If not provided, a new figure and axes will be created.

    Example:
    ---------
    >>> import pandas as pd
    >>> from statsmodels.api import OLS
    >>> from estyp.linear_model import Stepwise
    >>> data = pd.DataFrame({"y": [1,2,3,4,5], "x1": [5,20,3,2,1], "x2": [6,7,8,9,10]})
    >>> stepwise = Stepwise(formula="y ~ 1", data=data, model=OLS, direction="forward", criterion="aic")
    >>> stepwise.fit()

    Notes:
    ------
    - The class is designed to work seamlessly with statsmodels models.
    - If using "both" as the direction, the "f-test" criterion is not available.
    - Ensure that the data provided is appropriate for the model chosen.

    """

    def __init__(
        self,
        formula: str,
        data: DataFrame,
        model: Union[GLM, OLS, Logit, LogisticRegression] = OLS,
        direction: Literal["both", "forward", "backward"] = "forward",
        criterion: Literal["aic", "bic", "f-test"] = "aic",
        *,
        alpha: float = 0.05,
        max_iter: int = 100,
        formula_params: Dict[str, Any] = dict(),
        fit_params: Dict[str, Any] = {"disp": 0},
        verbose: bool = True,
    ):
        if criterion not in ["aic", "bic", "f-test"]:
            raise ValueError("criterion must be one of 'aic', 'bic', or 'f-test'")
        if not isinstance(formula, str):
            raise TypeError("formula must be a string")
        if not isinstance(data, DataFrame):
            raise TypeError("data must be a pandas DataFrame")
        if direction not in ["both", "forward", "backward"]:
            raise ValueError(
                "direction must be one of 'both', 'forward', or 'backward'"
            )
        if not isinstance(formula_params, dict):
            raise TypeError("formula_params must be a dictionary")
        if not isinstance(fit_params, dict):
            raise TypeError("fit_params must be a dictionary")
        if not (0 < alpha < 1):
            raise ValueError("alpha must be between 0 and 1")
        if direction == "both" and criterion == "f-test":
            raise ValueError(
                "direction 'both' is not available with criterion 'f-test'"
            )
        if not isinstance(max_iter, int):
            raise TypeError("max_iter must be an integer")
        if max_iter < 1:
            raise ValueError("max_iter must be greater than 0")

        self.formula = formula
        self.data = data
        self.model = model
        self.direction = direction
        self.criterion = criterion
        self.alpha = alpha
        self.max_iter = max_iter
        self.formula_params = formula_params
        self.fit_params = fit_params
        self.verbose = verbose

        self.__vr = self.formula.split("~")[0].strip()
        self.__data_vars = self.data.drop(columns=self.__vr).columns.tolist()
        if self.direction in ["backward", "both"]:
            preds = self.formula.split("~")[1].split("+")
            preds = [x for x in preds if x not in ["-1", "0"]]
        else:
            preds = self.__data_vars
        self.preds = tuple([x.strip() for x in preds])

        self.__metric_name = (
            "p-value" if self.criterion == "f-test" else self.criterion.upper()
        )

    def __forward_selection(self):
        m_actual = self.model.from_formula(
            self.formula, self.data, **self.formula_params
        ).fit(**self.fit_params)
        preds = self.formula.split("~")[1].split("+")
        preds = [x.strip() for x in preds]
        preds = [x for x in preds if x not in ["-1", "0"]]

        self.__BEST_METRICS = []

        termino = False
        cant = sum([1 for x in preds if x in self.data.columns])
        if cant == self.data.shape[1] - 1:
            termino = True
            MIN_METRIC = "Not calculated"

        if self.verbose and self.criterion in ["aic", "bic"]:
            print(
                f"Starting {self.__metric_name}: {m_actual.__getattribute__(self.criterion):0.4f}"
            )

        self.alpha = (
            m_actual.__getattribute__(self.criterion)
            if self.criterion in ["aic", "bic"]
            else self.alpha
        )
        remaining_preds = [x for x in self.preds if x not in self.formula]
        cant_preds = len(remaining_preds)

        f_actual = self.formula

        while not termino:
            METRICS = []
            for p in remaining_preds:
                f_prueba = f"{f_actual} + {p}"
                m_prueba = self.model.from_formula(
                    f_prueba, self.data, **self.formula_params
                ).fit(**self.fit_params)
                if self.criterion in ["aic", "bic"]:
                    METRIC = m_prueba.__getattribute__(self.criterion)
                else:
                    METRIC = nested_models_test(m_actual, m_prueba).p_value
                METRICS.append(METRIC)
            MIN_METRIC = min(METRICS)
            if MIN_METRIC >= self.alpha:
                self.__BEST_METRICS.append(MIN_METRIC)
                self.optimal_metric_ = MIN_METRIC
                MIN_METRIC = self.alpha
                termino = True
                m_actual = self.model.from_formula(
                    f_actual, self.data, **self.formula_params
                ).fit(**self.fit_params)
            else:
                self.alpha = (
                    MIN_METRIC if self.criterion in ["aic", "bic"] else self.alpha
                )
                var_min = remaining_preds[METRICS.index(MIN_METRIC)]
                f_actual = f"{f_actual} + {var_min}"
                options = [" 1 +", " + 1", "+ 1", "+1"]
                for option in options:
                    if option in f_actual:
                        f_actual = f_actual.replace(option, "")
                m_actual = self.model.from_formula(
                    f_actual, self.data, **self.formula_params
                ).fit(**self.fit_params)
                remaining_preds.remove(var_min)
                termino = not bool(remaining_preds)
                if self.verbose:
                    print(
                        f'- Term added: "{var_min}" | {self.__metric_name}: {MIN_METRIC:0.4f}'
                    )

        if self.verbose:
            mm_display = (
                f"{MIN_METRIC:0.4f}" if isinstance(MIN_METRIC, float) else MIN_METRIC
            )
            print(
                bcolors.OKGREEN
                + bcolors.UNDERLINE
                + bcolors.BOLD
                + "Forward selection completed"
                + bcolors.ENDC
            )
            if self.criterion in ["aic", "bic"]:
                print(f"- Obtained {self.__metric_name}: {mm_display}")
            print(
                "- Added terms:",
                cant_preds - len(remaining_preds) if remaining_preds else "None",
            )
            try:
                print(f'- Obtained formula: "{m_actual.model.formula}"')
            except:
                print(f'- Obtained formula: "{m_actual.formula}"')

        return m_actual

    def __backward_selection(self):
        m_actual = self.model.from_formula(
            self.formula, self.data, **self.formula_params
        ).fit(**self.fit_params)
        preds = self.formula.split("~")[1].split("+")
        preds = [x.strip() for x in preds]
        preds = [x for x in preds if x not in ["-1", "0", "1"]]
        self.__BEST_METRICS = []
        termino = False
        cant = len(preds)
        if cant == 0:
            termino = True
            MIN_METRIC = "Not calculated"

        if self.verbose and self.criterion in ["aic", "bic"]:
            print(
                f"Starting {self.__metric_name}: {m_actual.__getattribute__(self.criterion):0.4f}"
            )

        self.alpha = (
            m_actual.__getattribute__(self.criterion)
            if self.criterion in ["aic", "bic"]
            else -self.alpha
        )
        remaining_preds = [x for x in preds]
        cant_preds = len(remaining_preds)

        f_actual = self.formula

        only_one = False
        while not termino:
            METRICS = []
            for p in remaining_preds:
                p_prueba = [x for x in remaining_preds if x != p]
                f_prueba = f"{self.__vr} ~ {' + '.join(p_prueba) if len(p_prueba) != 0 else '1'}"
                if not only_one:
                    m_prueba = self.model.from_formula(
                        f_prueba, self.data, **self.formula_params
                    ).fit(**self.fit_params)
                else:
                    m_prueba = m_prueba
                if len(p_prueba) == 0:
                    only_one = True
                if self.criterion in ["aic", "bic"]:
                    METRIC = m_prueba.__getattribute__(self.criterion)
                else:
                    METRIC = -nested_models_test(m_prueba, m_actual).p_value
                METRICS.append(METRIC)
            MIN_METRIC = min(METRICS)
            if MIN_METRIC >= self.alpha:
                MIN_METRIC = self.alpha
                termino = True
                m_actual = self.model.from_formula(
                    f_actual, self.data, **self.formula_params
                ).fit(**self.fit_params)
            else:
                mm = MIN_METRIC if self.criterion != "f-test" else -MIN_METRIC
                self.__BEST_METRICS.append(mm)
                self.optimal_metric_ = mm
                self.alpha = (
                    MIN_METRIC if self.criterion in ["aic", "bic"] else self.alpha
                )
                var_min = remaining_preds[METRICS.index(MIN_METRIC)]
                remaining_preds.remove(var_min)
                f_actual = f"{self.__vr} ~ {' + '.join(remaining_preds) if remaining_preds else '1'}"
                options = [" 1 +", " + 1", "+ 1", "+1"]
                for option in options:
                    if option in f_actual:
                        f_actual = f_actual.replace(option, "")
                m_actual = self.model.from_formula(
                    f_actual, self.data, **self.formula_params
                ).fit(**self.fit_params)
                termino = not bool(remaining_preds)
                if self.verbose:
                    mm_display = (
                        -MIN_METRIC if self.criterion == "f-test" else MIN_METRIC
                    )
                    print(
                        f'- Term dropped: "{var_min}" | {self.__metric_name}: {mm_display:0.4f}'
                    )

        if self.verbose:
            mm_display = (
                f"{MIN_METRIC:0.4f}" if isinstance(MIN_METRIC, float) else MIN_METRIC
            )
            print(
                bcolors.OKGREEN
                + bcolors.UNDERLINE
                + bcolors.BOLD
                + "Backward selection completed"
                + bcolors.ENDC
            )
            if self.criterion in ["aic", "bic"]:
                print(f"- Obtained {self.__metric_name}: {mm_display}")
            print("- Dropped terms:", cant_preds - len(remaining_preds))
            try:
                print(f'- Obtained formula: "{m_actual.model.formula}"')
            except:
                print(f'- Obtained formula: "{m_actual.formula}"')

        return m_actual

    def __both_selection(self):
        fi = self.formula
        e = self.model.from_formula(fi, self.data, **self.formula_params)
        m = e.fit(**self.fit_params)
        preds = e.formula.split("+")
        data_vars = self.data.drop(columns=self.__vr).columns.tolist()
        preds[0] = preds[0].split("~")[1]
        preds = [x.strip() for x in preds]
        for x in ["0", "-1", "1"]:
            if x in preds:
                preds.remove(x)
        d = self.data.copy()
        data_preds = d.drop(columns=[self.__vr]).columns.tolist()
        METRICF = m.__getattribute__(self.criterion)
        self.__BEST_METRICS = [METRICF]
        METRIC0 = 0

        var_el = []
        var_ag = []

        if self.verbose:
            print(f"Starting {self.criterion.upper()}: {METRICF:0.4f}")

        for i in range(self.max_iter):
            if i == 0:
                METRICS = []
                if len(data_preds) != len(preds):
                    preds0 = preds.copy()
                    data0 = d.copy()
                    ff = f"{self.__vr} ~ {' + '.join(preds0) if preds0 else 1}"
                    continue
                for p in preds:
                    p0 = d.drop(columns=[p, self.__vr]).columns.tolist()
                    f0 = f"{self.__vr} ~ {' + '.join(p0) if p0 else 1}"
                    m0 = self.model.from_formula(f0, d, **self.formula_params).fit(
                        **self.fit_params
                    )
                    METRICS.append(m0.__getattribute__(self.criterion))
                METRIC0 = min(METRICS)
                id_min = METRICS.index(min(METRICS))
                if METRIC0 < METRICF:
                    self.__BEST_METRICS.append(METRIC0)
                    self.optimal_metric_ = METRIC0
                    action = "dropped"
                    var_min = preds[id_min]
                    data0 = d.drop(columns=var_min).copy()
                    preds0 = data0.drop(columns=self.__vr).columns.tolist()
                    ff = f"{self.__vr} ~ {' + '.join(preds0) if preds0 else 1}"
                    var_el.append(var_min)
                    m = self.model.from_formula(ff, d, **self.formula_params).fit(
                        **self.fit_params
                    )
                else:
                    if self.verbose:
                        print(
                            f"There are no improvements in {self.criterion.upper()}, so no terms will be dropped."
                        )
                    return m
            else:
                METRIC0 = METRICF
                METRICS_DROP = []
                for p in preds0:
                    p0 = data0.drop(columns=[p, self.__vr]).columns.tolist()
                    f0 = f"{self.__vr} ~ {' + '.join(p0) if p0 else 1}"
                    m0 = self.model.from_formula(f0, d, **self.formula_params).fit(
                        **self.fit_params
                    )
                    METRICS_DROP.append(m0.__getattribute__(self.criterion))
                METRICS_ADD = []
                for p in data_vars:
                    f0 = f"{ff} {('+ ' + p) if p not in preds0 else ''}"
                    m0 = (
                        self.model.from_formula(f0, d, **self.formula_params).fit(
                            **self.fit_params
                        )
                        if p not in preds0
                        else None
                    )
                    METRICS_ADD.append(
                        m0.__getattribute__(self.criterion)
                        if p not in preds0
                        else m.__getattribute__(self.criterion)
                    )
                MIN_ADD = min(METRICS_ADD)
                MIN_DROP = min(METRICS_DROP) if METRICS_DROP else MIN_ADD
                METRIC0 = MIN_ADD if MIN_ADD < MIN_DROP else MIN_DROP
                if METRIC0 < METRICF:
                    self.__BEST_METRICS.append(METRIC0)
                    self.optimal_metric_ = METRIC0
                    if MIN_DROP < MIN_ADD:
                        action = "dropped"
                        var_min = preds0[METRICS_DROP.index(MIN_DROP)]
                        data0.drop(columns=var_min, inplace=True)
                        preds0.remove(var_min)
                        ff = f"{self.__vr} ~ {' + '.join(preds0) if preds0 else 1}"
                        var_el.append(var_min)
                    else:
                        action = "added"
                        var_min = data_vars[METRICS_ADD.index(MIN_ADD)]
                        data0 = d[preds0 + [var_min, self.__vr]].copy()
                        preds0.append(var_min)
                        ff = f"{self.__vr} ~ {' + '.join(preds0) if preds0 else 1}"
                        var_ag.append(var_min)
                    m = self.model.from_formula(ff, d, **self.formula_params).fit(
                        **self.fit_params
                    )
                    METRICF = METRIC0
                else:
                    if self.verbose:
                        print(
                            bcolors.OKGREEN
                            + bcolors.UNDERLINE
                            + bcolors.BOLD
                            + "Both selection completed"
                            + bcolors.ENDC
                        )
                        print(f"- Obtained {self.criterion.upper()}: {METRICF:0.4f}")
                        print("- Dropped terms:", len(var_el))
                        print("- Added terms:", len(var_ag))
                        print(f'- Obtained formula: "{ff}"')
                    return m
            if self.verbose:
                print(
                    f'- Term "{var_min}" {action} | {self.criterion.upper()}: {METRIC0:0.4f}'
                )
            if not preds0:
                print(
                    bcolors.OKGREEN
                    + bcolors.UNDERLINE
                    + bcolors.BOLD
                    + "Both selection completed"
                    + bcolors.ENDC
                )
                print(f"- Obtained {self.criterion.upper()}: {METRICF:0.4f}")
                print("- Dropped terms:", len(var_el))
                print("- Added terms:", len(var_ag))
                print(f'- Obtained formula: "{ff}"')
                return m
        if self.verbose:
            print("Maximun number of iterations reached")
        return m

    def fit(self):
        """
        Conducts the stepwise process based on the specified direction and criterion.
        Once the stepwise process is completed, the optimal model, formula, and variables
        are stored as instance attributes.

        Examples:
        ---------
        >>> import pandas as pd
        >>> from statsmodels.api import OLS
        >>> from estyp.linear_model import Stepwise
        >>> data = pd.DataFrame({"y": [1,2,3,4,5], "x1": [5,20,3,2,1], "x2": [6,7,8,9,10]})
        >>> stepwise = Stepwise(formula="y ~ 1", data=data, model=OLS, direction="forward", criterion="aic")
        >>> stepwise.fit()
        >>> stepwise.optimal_variables_
        """
        if self.direction == "both":
            final_model = self.__both_selection()
        elif self.direction == "forward":
            final_model = self.__forward_selection()
        else:
            final_model = self.__backward_selection()

        self.optimal_model_ = final_model
        try:
            self.optimal_formula_ = final_model.model.formula
            self.optimal_variables_ = final_model.model.exog_names
        except:
            self.optimal_formula_ = final_model.formula
            self.optimal_variables_ = final_model.params.index.tolist()
        if "Intercept" in self.optimal_variables_:
            self.optimal_variables_.remove("Intercept")

    def plot_history(self, ax=None):
        """
        Plots the history of the chosen criterion (e.g., AIC, BIC, or F-test) during the stepwise process.
        This provides a visual representation of how the criterion changes as predictors are added or removed
        from the model.

        Parameters:
        -----------
        ax : matplotlib.axes.Axes, optional (default = None)
            An Axes instance for the plot. If not provided, a new figure and axes will be created.

        Returns:
        --------
        fig, ax : matplotlib.figure.Figure, matplotlib.axes.Axes
            The Figure and Axes instances containing the plot if not provided.

        Examples:
        ---------
        >>> import pandas as pd
        >>> from statsmodels.api import OLS
        >>> from estyp.linear_model import Stepwise
        >>> data = pd.DataFrame({"y": [1,2,3,4,5], "x1": [5,20,3,2,1], "x2": [6,7,8,9,10]})
        >>> stepwise = Stepwise(formula="y ~ x1 + x2", data=data, model=OLS, direction="forward", criterion="aic")
        >>> stepwise.fit()
        >>> fig, ax = stepwise.plot_history()"""
        if ax is not None and not isinstance(ax, plt.Axes):
            raise TypeError("ax must be an instance of matplotlib.axes.Axes")

        l = len(self.__BEST_METRICS)
        x = range(1, l + 1)

        if ax is None:
            fig, ax = plt.subplots()
            ax.plot(x, self.__BEST_METRICS, marker="o")
            ax.set_xticks(x)
            ax.set_xlabel("Iteration")
            ax.set_ylabel(self.criterion.upper())
            ax.set_title(
                f'Stepwise with direction="{self.direction}" and criterion="{self.criterion}"'
            )
            return fig, ax

        else:
            ax.plot(x, self.__BEST_METRICS, marker="o")
            ax.set_xticks(x)
            ax.set_xlabel("Iteration")
            ax.set_ylabel(self.criterion.upper())
            ax.set_title(
                f'Stepwise with direction="{self.direction}" and criterion="{self.criterion}"'
            )

    def __repr__(self) -> str:
        return f'Stepwise"(direction={self.direction}", criterion="{self.criterion}")'



from typing import Union

from pandas import DataFrame
from statsmodels.api import GLM, OLS, Logit

from estyp.linear_model import LogisticRegression
from estyp.linear_model.stepwise.__base import (__both_selection,
                                                __forward_selection)


def both_selection(
    formula: str,
    data: DataFrame,
    model: Union[GLM, OLS, Logit, LogisticRegression],
    *,
    max_iter: int = 10000,
    verbose: bool = False,
    formula_kwargs: dict = dict(),
    fit_kwargs:dict = {"disp": 0}
) -> str:
    """
[Deprecated] Both Forward and Backward Variable Selection for GLM's
======================================================

This function is deprecated. Use the `Stepwise()` class from the `estyp.linear_model` module instead.

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/linear_model.stepwise.html#both-method-variable-selection).

Description
-----------

This function performs both forward and backward variable selection using the Akaike Information Criterion (AIC).

Parameters
----------

- `formula`: A string representing the initial model formula.
- `data`: A Pandas DataFrame containing the data to be used for model fitting.
- `model`: A statsmodels.GLM object that represents the type of model to be fit.
- `max_iter`: The maximum number of iterations to perform.
- `**kwargs`: Additional keyword arguments to be passed to the model.from_formula() method.

Returns
-------

A string representing the final model formula.

Examples
--------

>>> import statsmodels.api as sm
>>> import pandas as pd
>>> data = pd.DataFrame({
    "y": [1, 2, 3, 4, 5],
    "x1": [1, 2, 3, 4, 5],
    "x2": [6, 7, 8, 9, 10],
})
>>> formula = "y ~ x1 + x2"
>>> model = sm.OLS
>>> final_formula = both_selection(formula=formula, data=data, model=model)
>>> print(final_formula)
    """
    return __both_selection(formula, data, model, max_iter, verbose, formula_kwargs, fit_kwargs)


def forward_selection(
    y: str,
    data: DataFrame,
    model: Union[GLM, OLS, Logit] = OLS,
    alpha: float = 0.05,
    *,
    verbose: bool = False,
    formula_kwargs: dict = dict(),
    fit_kwargs:dict = {"disp": 0}
):
    """
[Deprecated] Forward Variable Selection for GLM's
====================================

This function is deprecated. Use the `Stepwise()` class from the `estyp.linear_model` module instead.

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/linear_model.stepwise.html#forward-variable-selection-for-glm-s).

Description
-----------

This function performs forward variable selection using p-values calculated from nested models testing.

Parameters
----------

- `y`: A string containing the name of the dependent variable (target) to be predicted.
- data: The pandas DataFrame containing both the target variable 'y' and the        predictor variables for model training.
- `model`: A statsmodels model class. The statistical model to be used for model fitting and evaluation. Defaults to `sm.OLS`.
- `alpha`: A number between 0 and 1. The significance level for feature selection. A feature is added to the model if its p-value is less than this alpha value. Defaults to 0.05.
- `**kwargs`: Additional keyword arguments to be passed to the model.from_formula() method.

Returns
-------

A string representing the final model formula.

Examples
--------

>>> import pandas as pd
>>> import statsmodels.api as sm
>>> # Create sample DataFrame
>>> data = pd.DataFrame({
    'y': [1, 2, 3, 4, 5],
    'X1': [2, 4, 5, 7, 9],
    'X2': [3, 1, 6, 8, 4],
    'X3': [1, 5, 9, 2, 3]
})
>>> # Perform the forward variable selection
>>> formula = forward_selection(
    y = "y", ,
    data = data,
    model = sm.OLS,
    alpha = 0.05
)
>>> # Fit the model using the selected formula
>>> selected_model = sm.OLS.from_formula(formula, data).fit()
>>> print(selected_model.summary())
    """
    return __forward_selection(y, data, model, alpha, verbose, formula_kwargs, fit_kwargs)



from typing import List, Literal, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from pandas import Series, DataFrame
from statsmodels.regression.linear_model import RegressionResultsWrapper
from statsmodels.api import OLS

from estyp.testing.__base import (
    TestResults,
    _CheckHomocedasticity,
    _CheckIndependence,
    _CheckMulticollinearity,
    _CheckNormality,
    __nested_models_test,
    __prop_test,
    __t_test,
    __var_test,
    __cor_test,
    __chisq_test,
    _dw_test,
)


class CheckModel:
    """
Check Linear Regression Assumptions
===================================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#CheckModel).

Description
-----------

The `CheckModel` class provides methods to test the assumptions of the linear regression model.
These assumptions are:

- Normality of residuals
- Homoscedasticity (equal variance) of residuals
- Independence of residuals
- No Multicollinearity among predictors

Parameters
----------
`fitted_model` : `RegressionResultsWrapper`
    The fitted linear regression model which is an instance of RegressionResultsWrapper.

Example
-------

>>> import statsmodels.api as sm
>>> from sklearn.datasets import load_diabetes
>>> diabetes = load_diabetes()
>>> X = diabetes["data"]
>>> y = diabetes["target"]
>>> X = sm.add_constant(X)
>>> model = sm.OLS(y, X)
>>> fitted_model = model.fit()
>>> cm = CheckModel(fitted_model)
>>> cm.check_all()
    """

    def __init__(self, fitted_model: RegressionResultsWrapper):
        if not isinstance(fitted_model, RegressionResultsWrapper):
            raise TypeError(
                "The fitted model must be a RegressionResultsWrapper object"
            )
        self.model = fitted_model

    def check_normality(self, alpha=0.05, plot=True, return_pvals=False):
        """
Checks the normality assumption of the residuals using several statistical tests.

Parameters
----------
`alpha` : float, optional
    The significance level used by the statistical tests, by default 0.05.
`plot` : bool, optional
    If `True`, a plot is shown for visual inspection of normality, by default `True`.
`return_pvals` : bool, optional
    If `True`, the p-values of the statistical tests are returned, by default `False`.

Returns
-------
dict
    A dictionary of p-values of the statistical tests if `return_pvals=True`.
        """
        normality = _CheckNormality(self.model)
        if plot:
            fig, ax = normality.plot()
            plt.show()
        pvals = normality.test(sign_level=alpha)
        if return_pvals:
            return pvals

    def check_homocedasticity(self, alpha=0.05, plot=True, return_pvals=False):
        """
Checks the homoscedasticity assumption (equal variance of residuals) using several statistical tests.

Parameters
----------
`alpha` : float, optional
    The significance level used by the statistical tests, by default 0.05.
`plot` : bool, optional
    If `True`, a plot is shown for visual inspection of homoscedasticity, by default `True`.
`return_pvals` : bool, optional
    If `True`, the p-values of the statistical tests are returned, by default `False`.

Returns
-------
dict
    A dictionary of p-values of the statistical tests if `return_pvals=True`.
        """
        homocedasticity = _CheckHomocedasticity(self.model)
        if plot:
            fig, ax = homocedasticity.plot()
            plt.show()
        pvals = homocedasticity.test(sign_level=alpha)
        if return_pvals:
            return pvals

    def check_independence(self, alpha=0.05, plot=True, return_vals=False):
        """
Checks the independence assumption of the residuals using several statistical tests.

Parameters
----------
alpha : float, optional
    The significance level used by the statistical tests, by default 0.05.
plot : bool, optional
    If `True`, a plot is shown for visual inspection of independence, by default `True`.
return_pvals : bool, optional
    If `True`, the p-values of the statistical tests are returned, by default `False`.

Returns
-------
dict
    A dictionary of values of the statistical tests if `return_vals=True`.
        """
        independence = _CheckIndependence(self.model)
        if plot:
            fig, ax = independence.plot()
            plt.show()
        vals = independence.test(sign_level=alpha)
        if return_vals:
            return vals

    def check_multicollinearity(self, plot=True, return_cm=False):
        """
        Checks the multicollinearity assumption among predictors using the variance inflation factor (VIF).

        Parameters
        ----------
        `plot` : bool, optional
            If True, a plot is shown for visual inspection of multicollinearity, by default True.
        `return_cm` : bool, optional
            Returns the condition number of the model if True, by default False.

        Returns
        -------
        float
            The condition number of the model if `return_cm=True`.
        """
        multicollinearity = _CheckMulticollinearity(self.model)
        if plot:
            fig, ax = multicollinearity.plot()
            plt.show()
        pvals = multicollinearity.test()
        if return_cm:
            return pvals

    def check_all(self, alpha=0.05, plot=True, return_vals=False):
        """
Checks all the assumptions of the linear regression model.

Parameters
----------
`alpha` : float, optional
    The significance level used by the statistical tests, by default 0.05.
`plot` : bool, optional
    If `True`, a plot is shown for visual inspection of each assumption, by default `True`.
`return_vals` : bool, optional
    If `True`, the values of the statistical tests are returned, by default `False`.

Returns
-------
dict
    A dictionary of values of the statistical tests if `return_vals=True`.
        """
        normality = _CheckNormality(self.model)
        homocedasticity = _CheckHomocedasticity(self.model)
        independence = _CheckIndependence(self.model)
        multicollinearity = _CheckMulticollinearity(self.model)

        normality_results = normality.test(sign_level=alpha)
        if plot:
            fig, ax = normality.plot()
            plt.show()
        homocedasticity_results = homocedasticity.test(sign_level=alpha)
        if plot:
            fig, ax = homocedasticity.plot()
            plt.show()
        independence_results = independence.test(sign_level=alpha)
        if plot:
            fig, ax = independence.plot()
            plt.show()
        multicollinearity_results = multicollinearity.test()
        if plot:
            fig, ax = multicollinearity.plot()
            plt.show()
        names = ["Normality", "Homocedasticity", "Independence", "Multicollinearity"]
        tests = [
            normality_results,
            homocedasticity_results,
            independence_results,
            multicollinearity_results,
        ]
        if return_vals:
            return dict(zip(names, tests))


def var_test(
    x: Union[List, np.ndarray, Series],
    y: Union[List, np.ndarray, Series],
    *,
    ratio: Union[float, int] = 1,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    conf_level=0.95,
) -> TestResults:
    """
F Test to Compare Two Variances
===============================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#f-test-to-compare-two-variances).

Description
-----------
Performs an F test to compare the variances of two samples from normal populations. This function is inspired by the `var.test()` function of the software R.


Arguments
---------
- `x`, `y`: numeric list, `np.array` or `pd.Series` of data values.
- `ratio`: the hypothesized ratio of the population variances of x and y.
- `alternative`: a character string specifying the alternative hypothesis, must be one of "two-sided" (default), "greater" or "less". You can specify just the initial letter.
- `conf_level`: a number between 0 and 1 indicanting the confidence level of the interval.

Details
-------
The null hypothesis is that the ratio of the variances of the populations from which x and y were drawn, is equal to ratio.

Value
-----
An instance of the `TestResults` class containing the following attributes:

- `statistic`: the value of the F test statistic.
- `df`: the degrees of freedom for the F test statistic.
- `p_value`: the p-value for the test.
- `ci`: a confidence interval for the ratio of the population variances.
- `estimate`: the ratio of the sample variances of x and y.
- `alternative`: a string describing the alternative hypothesis.

Examples
--------

>>> import numpy as np
>>> np.random.seed(2023)
>>> x = np.random.normal(size=100)
>>> y = np.random.normal(size=100)

>>> print(var_test(x, y))
>>> print(var_test(x, y, alternative="less"))
>>> print(var_test(x, y, ratio=0.9, alternative="greater"))
    """
    return __var_test(x, y, ratio, alternative, conf_level)


def t_test(
    x: Union[List, np.ndarray, Series],
    y: Union[List, np.ndarray, Series] = None,
    *,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    mu: Union[float, int] = 0,
    paired=False,
    var_equal=False,
    conf_level=0.95,
) -> TestResults:
    """
Student's t-Test
================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#student-s-t-test).

Description
-----------
Performs one and two sample t-tests on groups of data. This function is inspired by the `t.test()` function of the software R.


Arguments
---------
- `x`: a (non-empty) numeric container of data values.
- `y`: an (optional) numeric container of data values.
- `alternative`: a string specifying the alternative hypothesis, must be one of "two-sided" (default), "greater" or "less".
- `mu`: a number indicating the true value of the mean (or difference in means if you are performing a two sample test).
- `paired`: a logical indicating whether you want a paired t-test.
- `var_equal`: a logical variable indicating whether to treat the two variances as being equal. If `True` then the pooled variance is used to estimate the variance otherwise the Welch (or Satterthwaite) approximation to the degrees of freedom is used.
- `conf_level`: a number between 0 and 1 indicanting the confidence level of the interval.

Details
-------
alternative = "greater" is the alternative that x has a larger mean than y. For the one-sample case: that the mean is positive.

If `paired` is `True` then both x and y must be specified and they must be the same length. Missing values are silently removed (in pairs if `paired` is `True`). If `var_equal` is `True` then the pooled estimate of the variance is used. By default, if `var_equal` is `False` then the variance is estimated separately for both groups and the Welch modification to the degrees of freedom is used.

Value
-----
An instance of the `TestResults` class containing the following attributes:

- `statistic`: the value of the t-statistic.
- `df`: the degrees of freedom for the t-statistic.
- `p_value`: the p-value for the test.
- `ci`: a confidence interval for the mean appropriate to the specified alternative hypothesis.
- `estimate`: the estimated mean or list of estimated means depending on whether it was a one-sample test or a two-sample test.
- `alternative`: a character string describing the alternative hypothesis.
- `mu`: the mean of the null hypothesis.

Examples
--------

>>> import numpy as np
>>> np.random.seed(2023)
>>> x = np.random.normal(size=100)
>>> y = np.random.normal(size=100)
>>> mu = 0.1

>>> print(t_test(x, mu=mu, alternative="less"))
>>> print(t_test(x, y, mu=mu))
>>> print(t_test(x, y, mu=mu, var_equal=True, alternative="greater"))
>>> print(t_test(x, y, mu=mu, paired=True))
    """
    return __t_test(x, y, alternative, mu, paired, var_equal, conf_level)


def nested_models_test(
    fitted_small_model: RegressionResultsWrapper,
    fitted_big_model: RegressionResultsWrapper,
) -> TestResults:
    """
Nested Models F-Test Function
=============================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#nested-models-f-test-function).

Description:
-------------
This function performs a nested models F-test using deviance from two fitted models from statsmodels library. The test compares two nested models: a larger or "big" model and a smaller or "small" model. The purpose of this test is to determine whether the larger model significantly improves the model fit compared to the smaller model by adding additional predictors.

Parameters:
-----------
`fitted_small_model` : `RegressionResultsWrapper`
    The fitted model representing the smaller/nested model. It has to come from statsmodels.
`fitted_big_model` : `RegressionResultsWrapper`
    The fitted model representing the larger model, which includes all the predictors from the smaller model and potentially additional predictors. It has to come from statsmodels.

Returns:
--------
The function returns an object of class TestResults that contains the following information:

`method` : str
    A string indicating the name of the statistical test (Nested models F-test).
`statistic` : float
    The computed F-statistic value.
`estimate` : float
    The difference in deviances between the models.
`df` : dict
    A dictionary with the degrees of freedom for the numerator and denominator of the F-statistic.
`p_value` : float
    The p-value associated with the F-statistic.

Examples:
---------
Example 1: With OLS

>>> import pandas as pd
>>> import statsmodels.api as sm
>>> data = pd.DataFrame({
...     "x": [2.01, 2.99, 4.01, 5.01, 6.89],
...     "y": [2, 3, 4, 5, 6]
... })
>>> model_small = sm.OLS.from_formula("y ~ 1", data).fit()
>>> model_big = sm.OLS.from_formula("y ~ x", data).fit()
>>> print(nested_models_test(model_small, model_big))

Example 2: With Logit

>>> import pandas as pd
>>> import statsmodels.api as sm
>>> data = pd.DataFrame({
...     "x": [2.01, 2.99, 4.01, 3.01, 4.89],
...     "y": [0, 1, 1, 0, 1]
... })
>>> model_small = sm.Logit.from_formula("y ~ 1", data).fit()
>>> model_big = sm.Logit.from_formula("y ~ x", data).fit()
>>> print(nested_models_test(model_small, model_big))

Example 3: With GLM

>>> import pandas as pd
>>> import statsmodels.api as sm
>>> data = pd.DataFrame({
...     "x": [2.01, 2.99, 4.01, 5.01, 6.89],
...     "y": [2, 3, 4, 5, 6]
... })
>>> model_small = sm.GLM.from_formula("y ~ 1", data, family = sm.families.Gamma()).fit()
>>> model_big = sm.GLM.from_formula("y ~ x", data, family = sm.families.Gamma()).fit()
>>> print(nested_models_test(model_small, model_big))
    """
    return __nested_models_test(fitted_small_model, fitted_big_model)


def prop_test(
    x: Union[List, np.ndarray, Series],
    *,
    n: Optional[Union[List, np.ndarray, Series]] = None,
    p: Optional[Union[float, List, np.ndarray]] = None,
    alternative: str = "two-sided",
    conf_level: float = 0.95,
    correct: bool = True,
) -> TestResults:
    """
Test of Equal or Given Proportions.
===================================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#test-of-equal-or-given-proportions).

Parameters
----------
`x` : array_like
    A vector of counts of successes, a one-dimensional table with two entries,
    or a two-dimensional table (or matrix) with 2 columns, giving the counts of
    successes and failures, respectively.
`n` : array_like, optional
    A vector of counts of trials; ignored if x is a matrix or a table.
    If not provided, it is calculated as the sum of the elements in x.
`p` : array_like, optional
    A vector of probabilities of success. The length of p must be the same as
    the number of groups specified by x, and its elements must be greater than
    0 and less than 1.
`alternative` : str, optional
    A character string specifying the alternative hypothesis, must be one of
    "two-sided" (default), "greater" or "less". You can specify just the
    initial letter. Only used for testing the null that a single proportion
    equals a given value, or that two proportions are equal; ignored otherwise.
`conf_level` : float, optional
    Confidence level of the returned confidence interval. Must be a single
    number between 0 and 1. Only used when testing the null that a single
    proportion equals a given value, or that two proportions are equal;
    ignored otherwise.
`correct` : bool, optional
    A logical indicating whether Yates' continuity correction should be
    applied where possible.

Returns
-------
`TestResults`
    A data class with the following attributes:

    - `statistic` : float
        The value of Pearson's chi-squared test statistic.
    - `df` : int
        The degrees of freedom of the approximate chi-squared distribution of
        the test statistic.
    - `p_value` : float
        The p-value of the test.
    - `estimate` : array_like
        A vector with the sample proportions x/n.
    - `null_value` : float or array_like
        The value of p if specified by the null hypothesis.
    - `conf_int` : array_like
        A confidence interval for the true proportion if there is one group,
        or for the difference in proportions if there are 2 groups and p is
        not given, or None otherwise. In the cases where it is not None, the
        returned confidence interval has an asymptotic confidence level as
        specified by conf_level, and is appropriate to the specified
        alternative hypothesis.
    - `alternative` : str
        A character string describing the alternative.
    - `method` : str
        A character string indicating the method used, and whether Yates'
        continuity correction was applied.

Examples
--------

>>> import numpy as np
>>> from scipy import stats
>>> x = np.array([83, 90, 129, 70])
>>> n = np.array([86, 93, 136, 82])
>>> result = prop_test(x, n=n)
>>> print(result)
    """
    return __prop_test(x, n, p, alternative, conf_level, correct)


def cor_test(
    x: Union[List, np.ndarray, Series],
    y: Union[List, np.ndarray, Series],
    *,
    method: Literal["pearson", "kendall", "spearman"] = "pearson",
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    conf_level: float = 0.95,
    continuity: bool = False,
) -> TestResults:
    """
Test for Association/Correlation Between Paired Samples
=======================================================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#test-for-association-correlation-between-paired-samples).

Description
-----------
Test for association between paired samples, using one of Pearson's product moment correlation coefficient, Kendall's
tau or Spearman's rho.

Arguments
---------
`x`, `y` : array_like
    Numeric one-dimensional `arrays`, `lists` or `pd.Series` of data values. `x` and `y` must have the same length.

`alternative` : str, optional
    Indicates the alternative hypothesis and must be one of "two-sided", "greater" or "less".
    "greater" corresponds to positive association, "less" to negative association.

`method` : str, optional
    A string indicating which correlation coefficient is to be used for the test.
    One of "pearson", "kendall", or "spearman".

`conf_level` : float, optional
    Confidence level for the returned confidence interval. Currently only used for the
    Pearson product moment correlation coefficient if there are at least 4 complete pairs of observations.

`continuity` : bool, optional
    If `True`, a continuity correction is used for Kendall's tau.

Returns
-------
A `TestResults` instance containing the following atributes:

- `statistic`: the value of the test statistic.
- `df` (if applicable): the degrees of freedom of the test statistic.
- `p_value`: the p-value of the test.
- `estimate`: the estimated measure of association.
- `null_value`: the value of the association measure under the null hypothesis, always 0.
- `alternative`: a string describing the alternative hypothesis.
- `method`: a string indicating how the association was measured.
- `conf_int` (if applicable): a confidence interval for the measure of association.

Details
-------
The three methods each estimate the association between paired samples and compute a test of the value being zero.
They use different measures of association, all in the range [-1, 1] with 0 indicating no association.
These are sometimes referred to as tests of no correlation, but that term is often confined to the default method.

References
----------
[1] D. J. Best & D. E. Roberts (1975). Algorithm AS 89: The Upper Tail Probabilities of Spearman's rho.
    Applied Statistics, 24, 377--379. 10.2307/2347111.

[2] Myles Hollander & Douglas A. Wolfe (1973), Nonparametric Statistical Methods.
    New York: John Wiley & Sons. Pages 185--194 (Kendall and Spearman tests).

Example
-------
Using the iris dataset to test the association between sepal length and petal length using Pearson's correlation:

>>> from sklearn import datasets
>>> iris = datasets.load_iris()
>>> sepal_length = iris.data[:, 0]
>>> petal_length = iris.data[:, 2]
>>> result = cor_test(sepal_length, petal_length, method="pearson")
>>> print(result)

    """
    return __cor_test(x, y, method, alternative, conf_level, continuity)


def chisq_test(
    x: Union[List, np.ndarray, Series],
    *,
    y: Union[List, np.ndarray, Series, None] = None,
    p: Union[List, np.ndarray, Series, None] = None,
    correct: bool = True,
    rescale_p: bool = False
):
    """
Pearson's Chi-squared Test for Count Data
=========================================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#pearson-s-chi-squared-test-for-count-data).

Description
-----------
`chisq_test()` performs chi-squared contingency table tests and goodness-of-fit tests.

Arguments
---------
`x`: arrat_like
    A numeric list or 2D list (matrix). x and y can also both be lists.
    
`y`: array_like
    A numeric data ; ignored if x is a matrix. If x is a list, y should be a list of the same length. The default is None.

`p`: array_like
    A list of probabilities of the same length as x. An error is raised if any entry of p is negative.

`correct`: 
    A boolean indicating whether to apply continuity correction when computing the test statistic for 2x2 tables: 
    one half is subtracted from all abs(O-E) differences; however, the correction will not be bigger than the differences themselves. The default is True.

`rescale_p`: boolean
    A boolean; if True then p is rescaled (if necessary) to sum to 1. If rescale_p is False, and p does not sum to 1, an error is raised.

Details
-------
    If x is a matrix with one row or column, or if x is a list and y is not given, then a goodness-of-fit test is performed 
    (x is treated as a one-dimensional contingency table). The entries of x must be non-negative integers. In this case, the hypothesis 
    tested is whether the population probabilities equal those in p, or are all equal if p is not given.

    If x is a matrix with at least two rows and columns, it is taken as a two-dimensional contingency table: the entries of x must be 
    non-negative integers. Otherwise, x and y must be lists of the same length; cases with None values are removed, the lists are 
    treated as factors, and the contingency table is computed from these. Then Pearson's chi-squared test is performed of the null 
    hypothesis that the joint distribution of the cell counts in a 2-dimensional contingency table is the product of the row and column marginals.

    The p-value is computed from the asymptotic chi-squared distribution of the test statistic; continuity correction is only used in 
    the 2-by-2 case (if correct is True, the default).

Returns
-------
    A `TestResults` instance containing the following attributes:

    `statistic`: 
        The value of the chi-squared test statistic.

    `df`: 
        The degrees of freedom of the approximate chi-squared distribution of the test statistic.

    `p_value`:
        The p-value for the test.

    `method`:
        A string indicating the type of test performed, and whether continuity correction was used.

    `expected`:
        The expected counts under the null hypothesis.


Examples
--------

From Agresti(2007) p.39

>>> M = [[762, 327, 468], [484, 239, 477]]
>>> result1 = chisq_test(M)
>>> print(result1)

Effect of rescale_p

>>> x = [12, 5, 7, 7]
>>> p = [0.4, 0.4, 0.2, 0.2]
>>> result2 = chisq_test(x, p=p, rescale_p=True)
>>> print(result2)

Testing for population probabilities

>>> x = [20, 15, 25]
>>> result31 = chisq_test(x)
>>> print(result31)

A second example of testing for population probabilities

>>> x = [89,37,30,28,2]
>>> p = [0.40,0.20,0.20,0.19,0.01]
>>> result32 = chisq_test(x, p=p)
>>> print(result32)

Goodness of fit

>>> x = [1, 2, 3, 4, 5, 6]
>>> y = [6, 1, 2, 3, 4, 5]
>>> result4 = chisq_test(x, y=y)
>>> print(result4)
    """

    return __chisq_test(x, y, p, correct, rescale_p)

def dw_test(
        input_obj: Union[str, OLS, RegressionResultsWrapper],
        *,
        order_by:Union[Series, np.ndarray] = None,
        alternative: Literal["two-sided", "greater", "less"] = "greater",
        data: DataFrame = None
) -> TestResults:
    """
Durbin-Watson Test
==================

View this in the [online documentation](https://estyp.readthedocs.io/en/latest/testing.html#durbin-watson-test-for-autocorrelation).

Description
-----------

Performs the Durbin-Watson test for autocorrelation of disturbances. Includes an approximated p-value taken from the `lmtest::dwtest()` function in R.


Parameters:
-----------

`input_obj`: One of str, OLS, or a fitted OLS object. 
        A formula description for the model to be tested (or a OLS object or a fitted OLS object).
`order_by`: Series, np.ndarray, optional
        The observations in the model are ordered by z. 
        If set to None (the default) the observations are assumed to be ordered (e.g., a time series).
`alternative`: str, optional.
        A character string specifying the alternative hypothesis. One of "greater" (default), "two-sided", or "less".
`data` : pd.DataFrame, optional 
        An optional pandas DataFrame containing the variables in the model.

Details
-------

The Durbin-Watson test has the null hypothesis that the autocorrelation of the disturbances is 0. 
It is possible to test against the alternative that it is greater than, not equal to, or less than 0, respectively. 
This can be specified by the alternative parameter.

For large sample sizes the algorithm might fail to compute the p value; in that case a warning is printed 
and an approximate p value will be given; this p value is computed using a normal approximation 
with mean and variance of the Durbin-Watson test statistic.

Returns
-------

A `TestResults` instance containing:

- `statistic`: the test statistic.
- `method`: a character string with the method used.
- `alternative`: a character string describing the alternative hypothesis.
- `p_value`: the corresponding p-value.
` `estimate`: the estimate of the autocorrelation of firt order of the residuals obtained from DW statistic.

References
----------

- J. Durbin & G.S. Watson (1950), Testing for Serial Correlation in Least Squares Regression I. Biometrika 37, 409–428.
- J. Durbin & G.S. Watson (1951), Testing for Serial Correlation in Least Squares Regression II. Biometrika 38, 159–177.
- J. Durbin & G.S. Watson (1971), Testing for Serial Correlation in Least Squares Regression III. Biometrika 58, 1–19.
- R.W. Farebrother (1980), Pan's Procedure for the Tail Probabilities of the Durbin-Watson Statistic. 
    Applied Statistics 29, 224–227.
- W. Krämer & H. Sonnberger (1986), The Linear Regression Model under Test. Heidelberg: Physica.

Examples
--------

Generate two AR(1) error terms with parameter rho = 0 (white noise) and rho = 0.9 respectively

- rho = 0:

>>> import numpy as np
>>> err1 = np.random.randn(100)
>>> ## generate regressor and dependent variable
>>> x = np.tile([-1, 1], 50)
>>> y1 = 1 + x + err1
>>> ## perform Durbin-Watson test
>>> dw_test('y1 ~ x', data=pd.DataFrame({'y1': y1, 'x': x}))

- rho = 0.9:

>>> from statsmodels.tsa.filters.filtertools import recursive_filter
>>> err2 = recursive_filter(err1, 0.9)
>>> y2 = 1 + x + err2
>>> dw_test('y2 ~ x', data=pd.DataFrame({'y2': y2, 'x': x}))
    """
    return _dw_test(input_obj, order_by, alternative, data)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name                          = "estyp",
    version                       = "0.9.0",
    author                        = "estebanrucan",
    author_email                  = "errucan@gmail.com",
    description                   = "Extended Statistical Toolkit Yet Practical",
    long_description              = readme,
    long_description_content_type = "text/markdown",
    url                           = "https://github.com/estebanrucan/estyp",
    project_urls                  = {
        "Bug Tracker": "https://github.com/estebanrucan/estyp/issues",
        "Homepage": "https://github.com/estebanrucan/estyp",
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages         = find_packages(),
    python_requires  = ">=3.9.12",
    install_requires = [
        "numpy >= 1.22.3",
        "scikit-learn >= 1.3.0",
        "matplotlib >= 3.4.3",
        "patsy >= 0.5.3",
        "statsmodels >= 0.13.5",
        "scipy >= 1.10.1",
        "kmodes >= 0.12.2",
        "tqdm >= 4.65.0"
    ],
    license              = "MIT",
    include_package_data = True,
)
